/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>
#include <string.h>

static int8_t compile_tmpdir_inode_structure(WOFS_STATE_ARG
	struct wofs_inode* const temp, const wofs_ino_t dir, const uint8_t shift)
{
	WOFS_READ_LOCAL(struct wofs_inode, inode,
		((wofs_addr_t)WOFS_INODE_TO_BLOCK(dir))<<shift);

	return wofs__compile_temp_inode(PASS_STATE temp, WOFS_LOCAL_ADDR(inode),
		WOFS_INOTYPE_TMPDIR);
}

static int8_t fill_temp_inode_except_type(WOFS_STATE_ARG
	const wofs_addr_t addr, const struct wofs_inode* const inode)
{
	/* remark: since 'i_type' and 'i_alive' is at the end of the inode */
	/* structure, and we have to leave 'i_type' as it is, we can just */
	/* write the first portion of the inode structure (and translate to */
	/* an error code) */

	if(0!=wroncefs_op_write(PASS_STATE inode, addr,
		offsetof(struct wofs_inode, i_type)))
	{
		return -WOFS_ERR_EIOW;
	}

	return 0;
}

static int8_t copy_non_dead_entries(WOFS_STATE_ARG
	const wofs_addr_t temp_addr, const wofs_ino_t dir,
	wofs_size_t* const size, const uint8_t shift)
{
	/* copy all non-dead entries (note: "alive" != "non-dead", we have */
	/* a special fsinfo directory entry in the root directory) */

	wofs_addr_t dst_addr = temp_addr + sizeof(struct wofs_inode);

	wofs_addr_t src_addr =
		(((wofs_addr_t)WOFS_INODE_TO_BLOCK(dir)) << shift) +
		sizeof(struct wofs_inode);

	const wofs_addr_t end_addr = src_addr + (((wofs_addr_t)1)<<shift) -
		/* we passed the end by 1 inode, correct it */
		sizeof(struct wofs_inode) -
		/* and directory ends with a directory entry */
		(sizeof(struct wofs_rawdirent) -
			sizeof(((struct wofs_rawdirent*)0)->d_name2)) -
		/* ... with at least 1 character long name (rounded up) */
		WOFS_ROUND_NAMELEN(1);

	char namebuf[WRONCEFS_NAME_MAX];

	while(src_addr<=end_addr)
	{
		WOFS_READ_LOCAL(struct wofs_rawdirent, rde, src_addr);

		/* d_ino is the last filled field, so stop if it is missing */
		if(WOFS_END_INODE==WOFS_LOCAL_FIELD(rde, d_ino))
		{
			break;
		}

		src_addr += sizeof(struct wofs_rawdirent) -
			sizeof(((struct wofs_rawdirent*)0)->d_name2);

		const uint8_t rounded_namelen =
			WOFS_ROUND_NAMELEN(WOFS_LOCAL_FIELD(rde, d_namelen));

		if(WOFS_DIRENT_DEAD==WOFS_LOCAL_FIELD(rde, d_alive))
		{
			src_addr += rounded_namelen;
			continue;
		}

		if(0!=wroncefs_op_write(PASS_STATE WOFS_LOCAL_ADDR(rde), dst_addr,
			sizeof(struct wofs_rawdirent)-
			sizeof(((struct wofs_rawdirent*)0)->d_name2)))
		{
			return -WOFS_ERR_EIOW;
		}

		dst_addr += sizeof(struct wofs_rawdirent) -
			sizeof(((struct wofs_rawdirent*)0)->d_name2);

		WOFS_READ_EXTERN(namebuf, src_addr, WOFS_LOCAL_FIELD(rde, d_namelen));

		if(0!=wroncefs_op_write(PASS_STATE namebuf, dst_addr,
			WOFS_LOCAL_FIELD(rde, d_namelen)))
		{
			return -WOFS_ERR_EIOW;
		}

		src_addr += rounded_namelen;
		dst_addr += rounded_namelen;
	}

	/* output data length */
	*size = dst_addr - (temp_addr + sizeof(struct wofs_inode));

	return 0;
}

WOFS_IMPL int8_t wofs__cleanup_directory(WOFS_STATE_ARG wofs_ino_t dir)
{
	int8_t ec;

	wofs_addr_t temp_addr;

	if(0!=(ec=wofs__get_temp_block(PASS_STATE &temp_addr)))
	{
		return ec;
	}

	const uint8_t shift = WOFS_STATE.blockszsh;

	struct wofs_inode temp;

	if(0!=(ec=compile_tmpdir_inode_structure(PASS_STATE &temp, dir, shift)))
	{
		return ec;
	}

	if(0!=(ec=fill_temp_inode_except_type(PASS_STATE temp_addr, &temp)))
	{
		return ec;
	}

	/* content here means: only the directory entries, without the inode */
	/* describing the block */
	wofs_size_t content_size;

	if(0!=(ec=copy_non_dead_entries(PASS_STATE temp_addr, dir, &content_size,
		shift)))
	{
	erase_and_exit:;
		/* try to fix it (ignore erase error) */
		wroncefs_op_erase(PASS_STATE temp_addr, ((wofs_size_t)1)<<shift);

		return ec;
	}

	/* we have a limit to length of the kept content */

	if(((((wofs_addr_t)1)<<shift)-sizeof(struct wofs_inode)-
		sizeof(struct wofs_tempdirinfo))<=content_size)
	{
		/* we must ensure this, although this shall not happen -- on the */
		/* other hand, how the heck did you got here? */
		ec = -WOFS_ERR_ENOTRECOVERABLE;
		goto erase_and_exit;
	}

	/* compile the temporary directory info, and write the first part */

	const wofs_addr_t tdi_addr = temp_addr + (((wofs_addr_t)1)<<shift) -
		sizeof(struct wofs_tempdirinfo);

	struct wofs_tempdirinfo tdi;

	tdi.orig_ino  = dir;
	tdi.cont_size = content_size;

	if(0!=wroncefs_op_write(PASS_STATE &tdi, tdi_addr, sizeof(tdi)))
	{
		ec = -WOFS_ERR_EIOW;
		goto erase_and_exit;
	}

	/* set the 'i_type' of our temporary inode */

	if(0!=wroncefs_op_write(PASS_STATE &temp.i_type,
		temp_addr+offsetof(struct wofs_inode, i_type), sizeof(temp.i_type)))
	{
		ec = -WOFS_ERR_EIOW;
		goto erase_and_exit;
	}

	/* erase the original inode, and use fill_1() */

	const wofs_addr_t orig_addr =
		((wofs_addr_t)WOFS_INODE_TO_BLOCK(dir)) << shift;

	/* restore the 'i_type', we will use this structure to fill the */
	/* original inode as well (rest of this structure is already correct) */
	temp.i_type = WOFS_INOTYPE_DIR;

	/*------------------- warning ---------------*/
	/* below this line, if any operation fails, control must go to */
	/* 'restore_failure' label */

	if(0!=wroncefs_op_erase(PASS_STATE orig_addr, ((wofs_size_t)1)<<shift))
	{
		ec = -WOFS_ERR_EIOE;
		goto restore_failure;
	}

	if(0!=(ec=wofs__ino_fill_1(PASS_STATE &temp, orig_addr)))
	{
		goto restore_failure;
	}

	/* copy back the content to the original inode */

	if(0!=(ec=wofs__copy_storage_content(PASS_STATE
		orig_addr+sizeof(struct wofs_inode),
		temp_addr+sizeof(struct wofs_inode), content_size)))
	{
		goto restore_failure;
	}

	/* fill_2() */

	if(0!=(ec=wofs__ino_fill_2(PASS_STATE &temp, orig_addr)))
	{
		goto restore_failure;
	}

	/* erase the temporary inode */

	if(0!=wroncefs_op_erase(PASS_STATE temp_addr, ((wofs_size_t)1)<<shift))
	{
		ec = -WOFS_ERR_EIOE;
		goto restore_failure;
	}

	/* done */

	return 0;

restore_failure:;
	// TODO: should we retry? mark filesystem? invoke a user callback?
	// (probably the best option: user callback + remount the filesystem)
	return ec;
}
