/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>
#include <string.h>

/* Fix a broken directory entry. */
/* NOTES: */
/* 1. 'd_namelen' is the last field which will be filled */
/* 2. 'd_namelen' and 'd_alive' are next to each other */
/* 3. since this will be the last "valid" (or not so valid) entry in the */
/*    current block, we can easily fix it by determine the name length (in */
/*    order to be able to skip it) and set 'd_alive' to zero */
static int8_t fix_broken_entry__(WOFS_STATE_ARG const wofs_ino_t ino,
	const uint8_t namelen, const uint8_t alive,
	const wofs_addr_t addr, const wofs_addr_t block_end)
{
	char d_name[WRONCEFS_NAME_MAX];
	struct wofs_rawdirent rde;

	if(0xFF==namelen)
	{
		/* name must be in the same block -- compute possible maximal len */
		size_t maxnamelen = block_end -
			(addr + (sizeof(struct wofs_rawdirent) -
			sizeof(((struct wofs_rawdirent*)0)->d_name2)));

		if(maxnamelen>WOFS_STATE.name_max)
		{
			maxnamelen = WOFS_STATE.name_max;
		}

		WOFS_READ_EXTERN(d_name, addr+(sizeof(struct wofs_rawdirent)-
			sizeof(((struct wofs_rawdirent*)0)->d_name2)), maxnamelen);
		char* ff = memchr(d_name, 0xFF, maxnamelen);

		if(NULL!=ff)
		{
			rde.d_namelen = ff - d_name;
		}
		else
		{
			rde.d_namelen = (uint8_t)maxnamelen;
		}

		if(0!=wroncefs_op_write(PASS_STATE &rde.d_namelen,
			addr+offsetof(struct wofs_rawdirent, d_namelen),
			sizeof(rde.d_namelen)))
		{
			return -WOFS_ERR_EIOW;
		}
	}

	if(0!=alive)
	{
		rde.d_alive = 0;

		if(0!=wroncefs_op_write(PASS_STATE &rde.d_alive,
			addr+offsetof(struct wofs_rawdirent, d_alive),
			sizeof(rde.d_alive)))
		{
			return -WOFS_ERR_EIOW;
		}
	}

	if(WOFS_END_INODE==ino)
	{
		rde.d_ino = WOFS_INVALID_INODE;

		if(0!=wroncefs_op_write(PASS_STATE &rde.d_ino,
			addr+offsetof(struct wofs_rawdirent, d_ino),
			sizeof(rde.d_ino)))
		{
			return -WOFS_ERR_EIOW;
		}
	}

	return 0;
}

WOFS_EXTERN void wofs__dir_open(WOFS_STATE_ARG struct wofs_dir* const dir,
	wofs_ino_t ino, const uint8_t namelen,
	struct wofs_direnthint* const out_hint)
{
	dir->dir_hint = out_hint;

	if(NULL!=out_hint)
	{
		out_hint->addr = 0;

		if(0<namelen && namelen<=WOFS_STATE.name_max)
		{
			/* hint tracking is valid */
			out_hint->namelen = WOFS_ROUND_NAMELEN(namelen);
			out_hint->ino     = ino;
			out_hint->found   = 0;
		}
		else
		{
			out_hint->ino = 0;
			dir->dir_hint = NULL;
		}
	}

	dir->dir_addr =
		((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << WOFS_STATE.blockszsh;

	dir->dir_end = 0;
	dir->dir_ino = ino;

	/* leave 'dir->dir_ent' unchanged -- semantically it is invalid, so */
	/* it's not our problem if the user reads the garbage from it */
}

WOFS_IMPL int8_t wofs__dir_read(WOFS_STATE_ARG struct wofs_dir* const dir)
{
	const uint8_t shift = WOFS_STATE.blockszsh;

	wofs_ino_t last_fix = WOFS_INVALID_INODE;

	for(;;)
	{
		WOFS_READ_LOCAL(struct wofs_inode, inode,
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(dir->dir_ino))<<shift);

		if(0==dir->dir_end)
		{
			dir->dir_end =
			/* The last possible valid directory entry is at the end of the */
			/* block. Minimal name length is 1 character, but it is aligned */
			/* up to 2 bytes. */
#define NAME_PRESERVE 2
			dir->dir_addr + (1 << shift) - (sizeof(struct wofs_rawdirent)
				- sizeof(((struct wofs_rawdirent*)0)->d_name2))
				- NAME_PRESERVE;

			dir->dir_addr += sizeof(struct wofs_inode);

			dir->dir_lost = 0;
		}

		const wofs_addr_t end = dir->dir_end;

		while(dir->dir_addr<=end)
		{
		refetch:;
			WOFS_READ_LOCAL(struct wofs_rawdirent, rde, dir->dir_addr);

			if(WOFS_END_INODE==WOFS_LOCAL_FIELD(rde, d_ino) &&
				0xFF==WOFS_LOCAL_FIELD(rde, d_name2)[0])
			{
				break;
			}

			dir->dir_addr += sizeof(struct wofs_rawdirent) -
				sizeof(((struct wofs_rawdirent*)0)->d_name2);

			if(WOFS_END_INODE==WOFS_LOCAL_FIELD(rde, d_ino) ||
				0xFF==WOFS_LOCAL_FIELD(rde, d_namelen))
			{
				/* this is a broken entry, which is a leftover of an */
				/* interrupted add_dirent() -- we try to fix it by */
				/* invalidating the entry */

				const int8_t ec =
					fix_broken_entry__(PASS_STATE
						WOFS_LOCAL_FIELD(rde, d_ino),
						WOFS_LOCAL_FIELD(rde, d_namelen),
						WOFS_LOCAL_FIELD(rde, d_alive),
						dir->dir_addr-(sizeof(struct wofs_rawdirent)-
						sizeof(((struct wofs_rawdirent*)0)->d_name2)),
						end+(sizeof(struct wofs_rawdirent)-
						sizeof(((struct wofs_rawdirent*)0)->d_name2))+
						NAME_PRESERVE);

				if(0!=ec)
				{
					return ec;
				}

				dir->dir_addr -= sizeof(struct wofs_rawdirent) -
					sizeof(((struct wofs_rawdirent*)0)->d_name2);
				goto refetch;
			}

			if(WOFS_STATE.name_max<WOFS_LOCAL_FIELD(rde, d_namelen))
			{
				return -WOFS_ERR_ENOTRECOVERABLE;
			}

			switch(WOFS_LOCAL_FIELD(rde, d_alive))
			{
			default:
				dir->dir_lost += (sizeof(struct wofs_rawdirent) -
					sizeof(((struct wofs_rawdirent*)0)->d_name2)) +
					WOFS_ROUND_NAMELEN(WOFS_LOCAL_FIELD(rde, d_namelen));
			case WOFS_DIRENT_FSINFO:
				dir->dir_addr +=
					WOFS_ROUND_NAMELEN(WOFS_LOCAL_FIELD(rde, d_namelen));
				continue;

			case WOFS_DIRENT_ALIVE:
				break;
			}

			WOFS_READ_EXTERN(dir->dir_ent.d_name, dir->dir_addr,
				WOFS_LOCAL_FIELD(rde, d_namelen));

			dir->dir_ent.d_namelen = WOFS_LOCAL_FIELD(rde, d_namelen);
			dir->dir_ent.d_ino     = WOFS_LOCAL_FIELD(rde, d_ino);

			dir->dir_addr +=
				WOFS_ROUND_NAMELEN(WOFS_LOCAL_FIELD(rde, d_namelen));

			/* we found a valid directory entry */
			return 0;
		}

		if(NULL!=dir->dir_hint && 0==dir->dir_hint->found)
		{
			dir->dir_hint->ino = dir->dir_ino;

			if(dir->dir_addr<=end &&
				dir->dir_hint->namelen<=((end-dir->dir_addr)+NAME_PRESERVE))
			{
				dir->dir_hint->addr = dir->dir_addr;
				/* stop hint tracking */
				dir->dir_hint->found = 1;
			}
		}

		if((WOFS_END_INODE!=WOFS_LOCAL_FIELD(inode, i_next) &&
			0x00!=WOFS_LOCAL_FIELD(inode, i_vnext)) || 0<dir->dir_lost)
		{
			/* chain extension was interrupted or there is lost space in */
			/* the directory block */

			if(last_fix==dir->dir_ino)
			{
				return -WOFS_ERR_ENOTRECOVERABLE;
			}

			const int8_t ec = wofs__cleanup_directory(PASS_STATE
				dir->dir_ino);

			if(0!=ec)
			{
				return ec;
			}

			last_fix = dir->dir_ino;

			/* reset 'dir_end' to refetch data, the block content has been */
			/* changed! */
			dir->dir_end = 0;

			/* directory cleanup can change 'dir_hint->addr', therefore */
			/* we need to reset it */

			if(NULL!=dir->dir_hint && (0==dir->dir_hint->found ||
				dir->dir_hint->ino==dir->dir_ino))
			{
				dir->dir_hint->found = 0;
				dir->dir_hint->addr  = 0;

				/* and rescan the current block */
				goto rescan_current_block;
			}

			continue;
		}

		dir->dir_ino = WOFS_LOCAL_FIELD(inode, i_next);

		if(0<dir->dir_ino && dir->dir_ino<=WOFS_STATE.nblocks)
		{
			const int8_t ec =
				wofs__check_inode(PASS_STATE dir->dir_ino, WOFS_INOTYPE_DIR);

			if(0!=ec)
			{
				return ec;
			}
		}
		else
		{
			break;
		}

	rescan_current_block:;

		dir->dir_addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(dir->dir_ino)) << shift;

		dir->dir_end = 0;
	}

	return -WOFS_ERR_ENOMORE;
}
