/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>

/*
	Layout to be generated:
		wofs_inode
		wofs_rawdirent
		wofs_fsinfo
		wofs_rawdirent
		".\xFF"
		wofs_rawdirent
		".."
*/

WOFS_IMPL int8_t wofs_format(WOFS_STATE_ARG uint16_t nblocks,
	uint8_t blockszsh, uint16_t flags)
{
	struct wofs_inode inode;

	/* prepare the root inode */

	inode.i_sig   = WOFS_SIG;
	inode.i_first = 1;
	inode.i_next  = WOFS_END_INODE;
	inode.i_vnext = 0xFF;
	inode.i_resd  = 0xFF;
	inode.i_type  = WOFS_INOTYPE_DIR;

	/* sanity check and erase everything */

	if(WOFS_MIN_NBLOCKS>nblocks || WOFS_MAX_NBLOCKS<nblocks)
	{
		return -WOFS_ERR_EINVAL;
	}

	if(WOFS_MIN_BLOCKSZSH>blockszsh || WOFS_MAX_BLOCKSZSH<blockszsh)
	{
		return -WOFS_ERR_EINVAL;
	}

	if(!(WOFS_FORMAT_DONTERASE & flags))
	{
		uint16_t blk;

		const uint16_t blksize = 1 << blockszsh;

		for(blk=0;blk<nblocks;++blk)
		{
			if(0!=wroncefs_op_erase(PASS_STATE ((wofs_addr_t)blk)<<blockszsh,
				blksize))
			{
				return -WOFS_ERR_EIOE;
			}
		}
	}

	/* things needed to be done: */
	/* fill inode: i_sig, i_first, i_type */
	/* fake directory entry (filesystem info) */
	/* "." and ".." */

	/* write the first part of the inode, and begin with a broken filesystem */
	{
		const int8_t ec = wofs__ino_fill_1(PASS_STATE &inode, 0);

		if(0!=ec)
		{
			return ec;
		}
	}

	wofs_addr_t addr = sizeof(struct wofs_inode);

	/* fake directory entry for fsinfo */
	{
		struct wofs_rawdirent de;

		de.d_ino     = WOFS_INVALID_INODE;
		de.d_namelen = sizeof(struct wofs_fsinfo);
		de.d_alive   = WOFS_DIRENT_FSINFO;

		if(0!=wroncefs_op_write(PASS_STATE &de, addr,
			sizeof(de)-sizeof(de.d_name2)))
		{
			return -WOFS_ERR_EIOW;
		}

		addr += sizeof(de) - sizeof(de.d_name2);
	}

	/* write filesystem info */
	{
		struct wofs_fsinfo fsinfo;

		fsinfo.fsi_blockszsh  = blockszsh;
		fsinfo.fsi_maxnamelen = WRONCEFS_NAME_MAX;
		fsinfo.fsi_nblocks    = nblocks;
		fsinfo.fsi_reserved   = 0xFFFF;
		fsinfo.fsi_flags      = wofs_fsi_flags;

		if(0!=wroncefs_op_write(PASS_STATE &fsinfo, addr, sizeof(fsinfo)))
		{
			return -WOFS_ERR_EIOW;
		}

		addr += sizeof(fsinfo);
	}

	{
		const int8_t ec = wofs__dflt_dir_layout(PASS_STATE addr,
			WOFS_ROOT_INODE, WOFS_ROOT_INODE);

		if(0!=ec)
		{
			return ec;
		}
	}

	/* write the remaining part of the root inode, and fix the broken */
	/* filesystem we made earlier */
	{
		const int8_t ec = wofs__ino_fill_2(PASS_STATE &inode, 0);

		if(0!=ec)
		{
			return ec;
		}
	}

	/* fill the state */

	WOFS_STATE.nblocks    = nblocks;
	WOFS_STATE.freeblocks = nblocks - 1 /* rootdir */ - WOFS_NUM_SPARE_BLOCKS;
	WOFS_STATE.blockszsh  = blockszsh;
	WOFS_STATE.name_max   = WRONCEFS_NAME_MAX;

	return 0;
}
