/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <string.h>
#include <stddef.h>

#define CALLBACK(_func_, _args_) \
	do \
	{ \
		if(NULL!=ops && NULL!=ops->_func_) \
		{ \
			ops->_func_ _args_; \
		} \
	} \
	while(0)

static int8_t fsck__inode_chain_has(WOFS_STATE_ARG const uint8_t blockszsh,
	wofs_ino_t ino, const wofs_ino_t subject, const uint8_t type,
	uint8_t* const found)
{
	int8_t ec;

	while(WOFS_INVALID_INODE!=ino && WOFS_END_INODE!=ino)
	{
		if(0!=wofs__check_inode(PASS_STATE ino, 0))
		{
			/* not a valid inode */
			break;
		}

		const wofs_addr_t addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh;

		WOFS_READ_LOCAL(uint8_t, i_type,
			addr+offsetof(struct wofs_inode, i_type));

		if(type!=WOFS_LOCAL_VALUE(i_type))
		{
			/* 'subject' is broken */
			break;
		}

		WOFS_READ_EXTERN(&ino, addr+offsetof(struct wofs_inode, i_next),
			sizeof(ino));

		if(subject==ino)
		{
			*found = 1;
			return 0;
		}
	}

	*found = 0;
	return 0;
}

static int8_t fsck__fix_inode_chains(WOFS_STATE_ARG const uint16_t nblocks,
	const uint8_t blockszsh, const struct wofs_fsckops* const ops)
{
	int8_t ec;

	for(wofs_ino_t ino=1;ino<=nblocks;++ino)
	{
		if(0!=wofs__check_inode(PASS_STATE ino, 0))
		{
			/* not a valid inode, skip */
			continue;
		}

		const wofs_addr_t addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh;

		WOFS_READ_LOCAL(uint8_t, i_type,
			addr+offsetof(struct wofs_inode, i_type));

		switch(WOFS_LOCAL_VALUE(i_type))
		{
		case WOFS_INOTYPE_FILE:
			{
				struct wofs_fd fd;

				if(0!=(ec=wofs__fd_setup(PASS_STATE &fd, ino)))
				{
					return ec;
				}

				WOFS_READ_LOCAL(struct wofs_inode, inode, addr);

				if(WOFS_END_INODE!=WOFS_LOCAL_FIELD(inode, i_next) &&
					0x00!=WOFS_LOCAL_FIELD(inode, i_vnext))
				{
					CALLBACK(fix_chain, (PASS_STATE ino));

					if(0!=(ec=wofs__file_rewrite(PASS_STATE &fd, fd.fd_clen)))
					{
						return ec;
					}
				}
			}
			break;

		case WOFS_INOTYPE_DIR:
			{
				WOFS_READ_LOCAL(struct wofs_inode, inode, addr);

				if(WOFS_END_INODE!=WOFS_LOCAL_FIELD(inode, i_next) &&
					0x00!=WOFS_LOCAL_FIELD(inode, i_vnext))
				{
					CALLBACK(fix_chain, (PASS_STATE ino));

					if(0!=(ec=wofs__cleanup_directory(PASS_STATE ino)))
					{
						return ec;
					}
				}
			}
			break;
		}
	}

	/* i_next/i_vnext fields are now corrected, perform inode link check */

	for(wofs_ino_t ino=1;ino<=nblocks;++ino)
	{
		if(0!=wofs__check_inode(PASS_STATE ino, 0))
		{
			/* not a valid inode, skip */
			continue;
		}

		const wofs_addr_t addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh;

		WOFS_READ_LOCAL(uint8_t, i_type,
			addr+offsetof(struct wofs_inode, i_type));

		if(WOFS_INOTYPE_FILE!=WOFS_LOCAL_VALUE(i_type) &&
			WOFS_INOTYPE_DIR!=WOFS_LOCAL_VALUE(i_type))
		{
			/* we interested in only files and directories */
			continue;
		}

		WOFS_READ_LOCAL(wofs_ino_t, i_first,
			addr+offsetof(struct wofs_inode, i_first));

		if(ino!=WOFS_LOCAL_VALUE(i_first))
		{
			uint8_t found;

			if(0!=(ec=fsck__inode_chain_has(PASS_STATE blockszsh,
				WOFS_LOCAL_VALUE(i_first), ino, WOFS_LOCAL_VALUE(i_type),
				&found)))
			{
				return ec;
			}

			if(!found)
			{
				CALLBACK(lost_inode, (PASS_STATE ino));

				if(wroncefs_op_erase(PASS_STATE
					((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh,
					((wofs_size_t)1) << blockszsh))
				{
					return -WOFS_ERR_EIOE;
				}
			}
		}
	}

	return 0;
}

/* Returns the index where it was found; or negative if not. */
static int16_t find_inode_in_array(const wofs_ino_t* inodes, int16_t count,
	const wofs_ino_t ino)
{
	for(int16_t i=0;i<count;++i)
	{
		if(ino==inodes[i])
		{
			return i;
		}
	}

	return -1;
}

/* Number of inodes to collect before running the directory scanner. Larger */
/* number means faster the scan, but it will use more memory. Must not be */
/* larger than 32767. */
#define WOFS_LOSTSCAN_WINDOWSIZE 12

static int8_t fsck__scan_and_erase(WOFS_STATE_ARG const uint16_t nblocks,
	const uint8_t blockszsh, const struct wofs_fsckops* const ops,
	wofs_ino_t* const inodes, size_t count)
{
	int8_t ec;

	for(wofs_ino_t ino=1;ino<=nblocks && 0<count;++ino)
	{
		if(0!=wofs__check_inode(PASS_STATE ino, WOFS_INOTYPE_DIR))
		{
			/* not a valid inode, skip */
			continue;
		}

		if(0<=find_inode_in_array(inodes, count, ino))
		{
			/* it is absolutely pointless to scan a (possibly) directory */
			/* inode whether it contains itself -- actually it does via the */
			/* "." entry, but it is ignored anyway */
			continue;
		}

		const wofs_addr_t addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh;

		WOFS_READ_LOCAL(wofs_ino_t, i_first,
			addr+offsetof(struct wofs_inode, i_first));

		/* check only complete directories */

		if(ino==WOFS_LOCAL_VALUE(i_first))
		{
			struct wofs_dir d;

			wofs__dir_open(PASS_STATE &d, ino, 0, NULL);

			while(0<count)
			{
				ec = wofs__dir_read(PASS_STATE &d);

				if((-WOFS_ERR_ENOMORE)==ec)
				{
					ec = 0;
					break;
				}
				else
				if(0!=ec)
				{
					return ec;
				}

				if(1==d.dir_ent.d_namelen && '.'==d.dir_ent.d_name[0])
				{
					/* skip "." */
					continue;
				}

				if(2==d.dir_ent.d_namelen && '.'==d.dir_ent.d_name[0] &&
					'.'==d.dir_ent.d_name[1])
				{
					/* skip ".." */
					continue;
				}

				int16_t idx = find_inode_in_array(inodes, count,
					d.dir_ent.d_ino);

				if(0<=idx)
				{
					inodes[idx] = inodes[--count];
				}
			}
		}
	}

	/* erase the leftover inodes: they were not referenced */

	for(int16_t i=0;i<count;++i)
	{
		CALLBACK(lost_inode, (PASS_STATE inodes[i]));

		if(wroncefs_op_erase(PASS_STATE
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(inodes[i])) << blockszsh,
			((wofs_size_t)1) << blockszsh))
		{
			return -WOFS_ERR_EIOE;
		}
	}

	return 0;
}

static int8_t fsck__scan_lost_inodes(WOFS_STATE_ARG const uint16_t nblocks,
	const uint8_t blockszsh, const struct wofs_fsckops* const ops)
{
	int8_t ec;

	wofs_ino_t inodes[WOFS_LOSTSCAN_WINDOWSIZE];
	int16_t count = 0;

	/* skip the root inode, it is always referenced implicitly */
	for(wofs_ino_t ino=WOFS_ROOT_INODE+1;ino<=nblocks;++ino)
	{
		if(0!=wofs__check_inode(PASS_STATE ino, 0))
		{
			/* not a valid inode, skip */
			continue;
		}

		/* at this point, it is adequate to have a valid inode (it will be */
		/* either DIR or FILE) */

		const wofs_addr_t addr =
			((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << blockszsh;

		WOFS_READ_LOCAL(wofs_ino_t, i_first,
			addr+offsetof(struct wofs_inode, i_first));

		/* only chain heads need to be checked */

		if(ino==WOFS_LOCAL_VALUE(i_first))
		{
			if(WOFS_LOSTSCAN_WINDOWSIZE==count)
			{
				if(0!=(ec=fsck__scan_and_erase(PASS_STATE nblocks, blockszsh,
					ops, inodes, count)))
				{
					return ec;
				}

				count = 0;
			}

			inodes[count++] = ino;
		}
	}

	if(0<count && 0!=(ec=fsck__scan_and_erase(PASS_STATE nblocks, blockszsh,
		ops, inodes, count)))
	{
		return ec;
	}

	return 0;
}

static int8_t fsck_core(WOFS_STATE_ARG uint16_t nblocks, uint8_t blockszsh,
	const struct wofs_fsckops* const ops)
{
	int8_t ec;

	if(0!=(ec=fsck__fix_inode_chains(PASS_STATE nblocks, blockszsh, ops)))
	{
		return ec;
	}

	if(0!=(ec=fsck__scan_lost_inodes(PASS_STATE nblocks, blockszsh, ops)))
	{
		return ec;
	}

	return 0;
}

WOFS_IMPL int8_t wofs_fsck(WOFS_STATE_ARG uint16_t nblocks, uint8_t blockszsh,
	const struct wofs_fsckops* const ops)
{
	int8_t ec;

	const uint16_t flags = WOFS_MOUNT_NORMAL;

	if(0!=(ec=wofs__mount_pre(PASS_STATE nblocks, blockszsh, flags)) ||
		0!=(ec=fsck_core(PASS_STATE nblocks, blockszsh, ops)) ||
		0!=(ec=wofs__mount_post(PASS_STATE nblocks, blockszsh, flags)))
	{
		memset(&WOFS_STATE, 0, sizeof(WOFS_STATE));
		return ec;
	}

	return 0;
}
