/*
Write-once filesystem

Copyright (c) 2022 Aron Barath
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the Author nor the names of contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Source code: svn://svn.repo.hu/wroncefs/trunk
Contact the author: aron-dev@mailbox.org
*/

#include "impl.h"
#include <stddef.h>
#include <string.h>

static wofs_addr_t ino_to_addr(WOFS_STATE_ARG wofs_ino_t ino)
{
	return ((wofs_addr_t)WOFS_INODE_TO_BLOCK(ino)) << WOFS_STATE.blockszsh;
}

static int8_t make_object_almost_deleted(WOFS_STATE_ARG const wofs_ino_t ino,
	wofs_addr_t addr)
{
	wofs_ino_t next;

	WOFS_READ_LOCAL(struct wofs_inode, inode, addr);

	next = WOFS_LOCAL_FIELD(inode, i_next);

	if(WOFS_END_INODE==next)
	{
	single_inode:;

		/* only a single node, mark it with WOFS_INVALID_INODE, according */
		/* to the specification */

		const wofs_ino_t new_next = WOFS_INVALID_INODE;

		if(0!=wroncefs_op_write(PASS_STATE &new_next,
			addr+offsetof(struct wofs_inode, i_next), sizeof(wofs_ino_t)))
		{
			return -WOFS_ERR_EIOW;
		}

		return 0;
	}

	/* we have a chain -- check whether 'i_vnext' is valid first */

	if(0x00!=WOFS_LOCAL_FIELD(inode, i_vnext))
	{
		switch(WOFS_LOCAL_FIELD(inode, i_type))
		{
		case WOFS_INOTYPE_FILE:
			{
				struct wofs_fd fd;

				int8_t ec;

				if(0!=(ec=wofs__fd_setup(PASS_STATE &fd, ino)))
				{
					return ec;
				}

				/* refetch possibly changed fields */
				WOFS_READ_LOCAL(uint16_t, i_next,
					fd.fd_addr+offsetof(struct wofs_inode, i_next));
				WOFS_READ_LOCAL(uint8_t, i_vnext,
					fd.fd_addr+offsetof(struct wofs_inode, i_vnext));

				if(WOFS_END_INODE!=WOFS_LOCAL_VALUE(i_next) &&
					0x00!=WOFS_LOCAL_VALUE(i_vnext) &&
					0!=(ec=wofs__file_rewrite(PASS_STATE &fd, fd.fd_clen)))
				{
					return ec;
				}
			}
			break;

		case WOFS_INOTYPE_DIR:
			{
				const int8_t ec = wofs__cleanup_directory(PASS_STATE ino);

				if(0!=ec)
				{
					return ec;
				}
			}
			break;

		default:
			return -WOFS_ERR_EBADINODE;
		}

		/* chainging is fixed now, recheck */
		{
			WOFS_READ_LOCAL(uint16_t, i_next,
				addr+offsetof(struct wofs_inode, i_next));

			if(WOFS_END_INODE==WOFS_LOCAL_VALUE(i_next))
			{
				goto single_inode;
			}
		}
	}

	const uint8_t dead_mark = 0;

	while(WOFS_END_INODE!=next)
	{
		addr = ino_to_addr(PASS_STATE next);

		WOFS_READ_LOCAL(struct wofs_inode, inode, addr);

		next = WOFS_LOCAL_FIELD(inode, i_next);

		if(0!=wroncefs_op_write(PASS_STATE &dead_mark,
			addr+offsetof(struct wofs_inode, i_alive), sizeof(dead_mark)))
		{
			return -WOFS_ERR_EIOW;
		}
	}

	return 0;
}

static int8_t erase_object_chain(WOFS_STATE_ARG wofs_addr_t addr)
{
	wofs_ino_t next;

	const wofs_size_t block_size = ((wofs_size_t)1) << WOFS_STATE.blockszsh;

	{
		WOFS_READ_LOCAL(struct wofs_inode, inode, addr);

		next = WOFS_LOCAL_FIELD(inode, i_next);

		if(0!=wroncefs_op_erase(PASS_STATE addr, block_size))
		{
			return -WOFS_ERR_EIOE;
		}

		++WOFS_STATE.freeblocks;

		if(WOFS_INVALID_INODE==next)
		{
			/* we had only a single node -- marked with WOFS_INVALID_INODE */
			return 0;
		}
	}

	/* we have a chain */

	while(WOFS_END_INODE!=next)
	{
		addr = ino_to_addr(PASS_STATE next);

		WOFS_READ_LOCAL(struct wofs_inode, inode, addr);

		next = WOFS_LOCAL_FIELD(inode, i_next);

		if(0!=wroncefs_op_erase(PASS_STATE addr, block_size))
		{
			return -WOFS_ERR_EIOE;
		}

		++WOFS_STATE.freeblocks;
	}

	return 0;
}

WOFS_IMPL int8_t wofs__remove_object(WOFS_STATE_ARG wofs_ino_t ino,
	const uint8_t namelen, const char* const name,
	int8_t (*chk_rm_callback)(WOFS_STATE_ARG wofs_ino_t ino))
{
	struct wofs_dir d;
	int8_t ec;

	/* find the entry to remove */

	wofs__dir_open(PASS_STATE &d, ino, 0, NULL);

	for(;;)
	{
		const int8_t ec = wofs__dir_read(PASS_STATE &d);

		if((-WOFS_ERR_ENOMORE)==ec)
		{
			return -WOFS_ERR_ENOENT;
		}
		else
		if(0!=ec)
		{
			return ec;
		}

		if(namelen==d.dir_ent.d_namelen &&
			0==memcmp(name, d.dir_ent.d_name, namelen))
		{
			/* got it */
			break;
		}
	}

	const wofs_ino_t trg = d.dir_ent.d_ino;

	if(0!=(ec=chk_rm_callback(PASS_STATE trg)))
	{
		return ec;
	}

	/* at this point, 'd.dir_addr' is advanced to the next entry, move */
	/* it back to the start address of our find */

	const wofs_addr_t ent_addr = d.dir_addr - (sizeof(struct wofs_rawdirent) -
		sizeof(((struct wofs_rawdirent*)0)->d_name2)) -
		WOFS_ROUND_NAMELEN(d.dir_ent.d_namelen);

	const wofs_addr_t ino_addr = ino_to_addr(PASS_STATE trg);

	/* step 1 */

	{
		const uint8_t dead_mark = WOFS_DIRENT_DEAD;

		if(0!=wroncefs_op_write(PASS_STATE &dead_mark,
			ent_addr+offsetof(struct wofs_rawdirent, d_alive),
			sizeof(dead_mark)))
		{
			return -WOFS_ERR_EIOW;
		}
	}

	/* step 2 */

	if(0!=(ec=make_object_almost_deleted(PASS_STATE trg, ino_addr)))
	{
		return ec;
	}

	/* step 3 */

	if(0!=(ec=wofs__cleanup_directory(PASS_STATE d.dir_ino)))
	{
		return ec;
	}

	/* step 4 */

	{
		const uint8_t dead_mark = 0;

		if(0!=wroncefs_op_write(PASS_STATE &dead_mark,
			ino_addr+offsetof(struct wofs_inode, i_alive), sizeof(dead_mark)))
		{
			return -WOFS_ERR_EIOW;
		}
	}

	/* step 5 */

	if(0!=(ec=erase_object_chain(PASS_STATE ino_addr)))
	{
		return ec;
	}

	return 0;
}
